# MIT License
#
# Copyright (c) 2020 Brockmann Consult GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

DEFAULT_NUM_RETRIES = 200
DEFAULT_TILE_SIZE = 1000
DEFAULT_CRS = 'http://www.opengis.net/def/crs/EPSG/0/4326'
DEFAULT_TIME_TOLERANCE = '10M'  # 10 minutes

CDS_MAX_IMAGE_SIZE = 2500

CDS_DATA_STORE_ID = 'cds'
CDS_DATA_OPENER_ID = 'dataset:netcdf:cds'

# This table maps request parameter names to NetCDF variable names, and
# was generated by the following process:
#
# 1. Obtain the complete list of valid request parameters via the Web interface
#    by selecting every box and copying the parameter names out of the generated
#    API request.
# 2. For each request parameter, make a separate API request containing only
#    that request parameter, producing a NetCDF file containing only the
#    corresponding output parameter.
# 3. Read the name of the single output variable from the NetCDF file and
#    collate it with the original request parameter. (Also read the long_name
#    and units attributes.)
#
# In this way we are guaranteed to get the correct NetCDF variable name for
# each request parameter, without having to trust that the documentation is
# correct.

ERA5_PARAMETERS = (
    # 1. request parameter name in CDS API
    # 2. NetCDF variable name (NB: not always CF-conformant)
    # 3. units from NetCDF attributes
    # 4. "long name" from NetCDF attributes
    ('100m_u_component_of_wind', 'u100', 'm s**-1',
     '100 metre U wind component'),
    ('100m_v_component_of_wind', 'v100', 'm s**-1',
     '100 metre V wind component'),
    ('10m_u_component_of_neutral_wind', 'u10n', 'm s**-1',
     'Neutral wind at 10 m u-component'),
    ('10m_u_component_of_wind', 'u10', 'm s**-1', '10 metre U wind component'),
    ('10m_v_component_of_neutral_wind', 'v10n', 'm s**-1',
     'Neutral wind at 10 m v-component'),
    ('10m_v_component_of_wind', 'v10', 'm s**-1', '10 metre V wind component'),
    ('10m_wind_speed', 'si10', 'm s**-1', '10 metre wind speed'),
    ('2m_dewpoint_temperature', 'd2m', 'K', '2 metre dewpoint temperature'),
    ('2m_temperature', 't2m', 'K', '2 metre temperature'),
    ('air_density_over_the_oceans', 'p140209', 'kg m**-3',
     'Air density over the oceans'),
    ('angle_of_sub_gridscale_orography', 'anor', 'radians',
     'Angle of sub-gridscale orography'),
    ('anisotropy_of_sub_gridscale_orography', 'isor', '~',
     'Anisotropy of sub-gridscale orography'),
    ('benjamin_feir_index', 'bfi', 'dimensionless', 'Benjamin-Feir index'),
    ('boundary_layer_dissipation', 'bld', 'J m**-2',
     'Boundary layer dissipation'),
    ('boundary_layer_height', 'blh', 'm', 'Boundary layer height'),
    ('charnock', 'chnk', '~', 'Charnock'),
    ('clear_sky_direct_solar_radiation_at_surface', 'cdir', 'J m**-2',
     'Clear-sky direct solar radiation at surface'),
    ('cloud_base_height', 'cbh', 'm', 'Cloud base height'),
    ('coefficient_of_drag_with_waves', 'cdww', 'dimensionless',
     'Coefficient of drag with waves'),
    ('convective_available_potential_energy', 'cape', 'J kg**-1',
     'Convective available potential energy'),
    ('convective_inhibition', 'cin', 'J kg**-1', 'Convective inhibition'),
    ('convective_precipitation', 'cp', 'm', 'Convective precipitation'),
    ('convective_rain_rate', 'crr', 'kg m**-2 s**-1', 'Convective rain rate'),
    ('convective_snowfall', 'csf', 'm of water equivalent',
     'Convective snowfall'),
    ('convective_snowfall_rate_water_equivalent', 'csfr', 'kg m**-2 s**-1',
     'Convective snowfall rate water equivalent'),
    ('downward_uv_radiation_at_the_surface', 'uvb', 'J m**-2',
     'Downward UV radiation at the surface'),
    ('duct_base_height', 'dctb', 'm', 'Duct base height'),
    ('eastward_gravity_wave_surface_stress', 'lgws', 'N m**-2 s',
     'Eastward gravity wave surface stress'),
    ('eastward_turbulent_surface_stress', 'ewss', 'N m**-2 s',
     'Eastward turbulent surface stress'),
    ('evaporation', 'e', 'm of water equivalent', 'Evaporation'),
    ('forecast_albedo', 'fal', '(0 - 1)', 'Forecast albedo'),
    ('forecast_logarithm_of_surface_roughness_for_heat', 'flsr', '~',
     'Forecast logarithm of surface roughness for heat'),
    ('forecast_surface_roughness', 'fsr', 'm', 'Forecast surface roughness'),
    ('free_convective_velocity_over_the_oceans', 'p140208', 'm s**-1',
     'Free convective velocity over the oceans'),
    ('friction_velocity', 'zust', 'm s**-1', 'Friction velocity'),
    ('gravity_wave_dissipation', 'gwd', 'J m**-2', 'Gravity wave dissipation'),
    ('high_cloud_cover', 'hcc', '(0 - 1)', 'High cloud cover'),
    ('high_vegetation_cover', 'cvh', '(0 - 1)', 'High vegetation cover'),
    ('ice_temperature_layer_1', 'istl1', 'K', 'Ice temperature layer 1'),
    ('ice_temperature_layer_2', 'istl2', 'K', 'Ice temperature layer 2'),
    ('ice_temperature_layer_3', 'istl3', 'K', 'Ice temperature layer 3'),
    ('ice_temperature_layer_4', 'istl4', 'K', 'Ice temperature layer 4'),
    ('instantaneous_10m_wind_gust', 'i10fg', 'm s**-1',
     'Instantaneous 10 metre wind gust'),
    ('instantaneous_eastward_turbulent_surface_stress', 'iews', 'N m**-2',
     'Instantaneous eastward turbulent surface stress'),
    ('instantaneous_large_scale_surface_precipitation_fraction', 'ilspf',
     '(0 - 1)', 'Instantaneous large-scale surface precipitation fraction'),
    ('instantaneous_moisture_flux', 'ie', 'kg m**-2 s**-1',
     'Instantaneous moisture flux'),
    ('instantaneous_northward_turbulent_surface_stress', 'inss', 'N m**-2',
     'Instantaneous northward turbulent surface stress'),
    ('instantaneous_surface_sensible_heat_flux', 'ishf', 'W m**-2',
     'Instantaneous surface sensible heat flux'),
    ('k_index', 'kx', 'K', 'K index'),
    ('lake_bottom_temperature', 'lblt', 'K', 'Lake bottom temperature'),
    ('lake_cover', 'cl', '(0 - 1)', 'Lake cover'),
    ('lake_depth', 'dl', 'm', 'Lake depth'),
    ('lake_ice_depth', 'licd', 'm', 'Lake ice depth'),
    ('lake_ice_temperature', 'lict', 'K', 'Lake ice temperature'),
    ('lake_mix_layer_depth', 'lmld', 'm', 'Lake mix-layer depth'),
    ('lake_mix_layer_temperature', 'lmlt', 'K', 'Lake mix-layer temperature'),
    ('lake_shape_factor', 'lshf', 'dimensionless', 'Lake shape factor'),
    ('lake_total_layer_temperature', 'ltlt', 'K',
     'Lake total layer temperature'),
    ('land_sea_mask', 'lsm', '(0 - 1)', 'Land-sea mask'),
    ('large_scale_precipitation', 'lsp', 'm', 'Large-scale precipitation'),
    ('large_scale_precipitation_fraction', 'lspf', 's',
     'Large-scale precipitation fraction'),
    ('large_scale_rain_rate', 'lsrr', 'kg m**-2 s**-1',
     'Large scale rain rate'),
    ('large_scale_snowfall', 'lsf', 'm of water equivalent',
     'Large-scale snowfall'),
    ('large_scale_snowfall_rate_water_equivalent', 'lssfr', 'kg m**-2 s**-1',
     'Large scale snowfall rate water equivalent'),
    ('leaf_area_index_high_vegetation', 'lai_hv', 'm**2 m**-2',
     'Leaf area index, high vegetation'),
    ('leaf_area_index_low_vegetation', 'lai_lv', 'm**2 m**-2',
     'Leaf area index, low vegetation'),
    ('low_cloud_cover', 'lcc', '(0 - 1)', 'Low cloud cover'),
    ('low_vegetation_cover', 'cvl', '(0 - 1)', 'Low vegetation cover'),
    ('magnitude_of_turbulent_surface_stress', 'magss', 'N m**-2 s',
     'Magnitude of turbulent surface stress'),
    ('maximum_individual_wave_height', 'hmax', 'm',
     'Maximum individual wave height'),
    ('mean_boundary_layer_dissipation', 'mbld', 'W m**-2',
     'Mean boundary layer dissipation'),
    ('mean_convective_precipitation_rate', 'mcpr', 'kg m**-2 s**-1',
     'Mean convective precipitation rate'),
    ('mean_convective_snowfall_rate', 'mcsr', 'kg m**-2 s**-1',
     'Mean convective snowfall rate'),
    ('mean_direction_of_total_swell', 'mdts', 'degrees',
     'Mean direction of total swell'),
    ('mean_direction_of_wind_waves', 'mdww', 'degrees',
     'Mean direction of wind waves'),
    ('mean_eastward_gravity_wave_surface_stress', 'megwss', 'N m**-2',
     'Mean eastward gravity wave surface stress'),
    ('mean_eastward_turbulent_surface_stress', 'metss', 'N m**-2',
     'Mean eastward turbulent surface stress'),
    ('mean_evaporation_rate', 'mer', 'kg m**-2 s**-1', 'Mean evaporation rate'),
    ('mean_gravity_wave_dissipation', 'mgwd', 'W m**-2',
     'Mean gravity wave dissipation'),
    ('mean_large_scale_precipitation_fraction', 'mlspf', 'Proportion',
     'Mean large-scale precipitation fraction'),
    ('mean_large_scale_precipitation_rate', 'mlspr', 'kg m**-2 s**-1',
     'Mean large-scale precipitation rate'),
    ('mean_large_scale_snowfall_rate', 'mlssr', 'kg m**-2 s**-1',
     'Mean large-scale snowfall rate'),
    ('mean_magnitude_of_turbulent_surface_stress', 'mmtss', 'N m**-2',
     'Mean magnitude of turbulent surface stress'),
    ('mean_northward_gravity_wave_surface_stress', 'mngwss', 'N m**-2',
     'Mean northward gravity wave surface stress'),
    ('mean_northward_turbulent_surface_stress', 'mntss', 'N m**-2',
     'Mean northward turbulent surface stress'),
    ('mean_period_of_total_swell', 'mpts', 's', 'Mean period of total swell'),
    ('mean_period_of_wind_waves', 'mpww', 's', 'Mean period of wind waves'),
    ('mean_potential_evaporation_rate', 'mper', 'kg m**-2 s**-1',
     'Mean potential evaporation rate'),
    ('mean_runoff_rate', 'mror', 'kg m**-2 s**-1', 'Mean runoff rate'),
    ('mean_sea_level_pressure', 'msl', 'Pa', 'Mean sea level pressure'),
    ('mean_snow_evaporation_rate', 'mser', 'kg m**-2 s**-1',
     'Mean snow evaporation rate'),
    ('mean_snowfall_rate', 'msr', 'kg m**-2 s**-1', 'Mean snowfall rate'),
    ('mean_snowmelt_rate', 'msmr', 'kg m**-2 s**-1', 'Mean snowmelt rate'),
    ('mean_square_slope_of_waves', 'msqs', 'dimensionless',
     'Mean square slope of waves'),
    ('mean_sub_surface_runoff_rate', 'mssror', 'kg m**-2 s**-1',
     'Mean sub-surface runoff rate'),
    ('mean_surface_direct_short_wave_radiation_flux', 'msdrswrf', 'W m**-2',
     'Mean surface direct short-wave radiation flux'),
    ('mean_surface_direct_short_wave_radiation_flux_clear_sky', 'msdrswrfcs',
     'W m**-2', 'Mean surface direct short-wave radiation flux, clear sky'),
    ('mean_surface_downward_long_wave_radiation_flux', 'msdwlwrf', 'W m**-2',
     'Mean surface downward long-wave radiation flux'),
    ('mean_surface_downward_long_wave_radiation_flux_clear_sky', 'msdwlwrfcs',
     'W m**-2', 'Mean surface downward long-wave radiation flux, clear sky'),
    ('mean_surface_downward_short_wave_radiation_flux', 'msdwswrf', 'W m**-2',
     'Mean surface downward short-wave radiation flux'),
    ('mean_surface_downward_short_wave_radiation_flux_clear_sky', 'msdwswrfcs',
     'W m**-2', 'Mean surface downward short-wave radiation flux, clear sky'),
    ('mean_surface_downward_uv_radiation_flux', 'msdwuvrf', 'W m**-2',
     'Mean surface downward UV radiation flux'),
    ('mean_surface_latent_heat_flux', 'mslhf', 'W m**-2',
     'Mean surface latent heat flux'),
    ('mean_surface_net_long_wave_radiation_flux', 'msnlwrf', 'W m**-2',
     'Mean surface net long-wave radiation flux'),
    ('mean_surface_net_long_wave_radiation_flux_clear_sky', 'msnlwrfcs',
     'W m**-2', 'Mean surface net long-wave radiation flux, clear sky'),
    ('mean_surface_net_short_wave_radiation_flux', 'msnswrf', 'W m**-2',
     'Mean surface net short-wave radiation flux'),
    ('mean_surface_net_short_wave_radiation_flux_clear_sky', 'msnswrfcs',
     'W m**-2', 'Mean surface net short-wave radiation flux, clear sky'),
    ('mean_surface_runoff_rate', 'msror', 'kg m**-2 s**-1',
     'Mean surface runoff rate'),
    ('mean_surface_sensible_heat_flux', 'msshf', 'W m**-2',
     'Mean surface sensible heat flux'),
    ('mean_top_downward_short_wave_radiation_flux', 'mtdwswrf', 'W m**-2',
     'Mean top downward short-wave radiation flux'),
    ('mean_top_net_long_wave_radiation_flux', 'mtnlwrf', 'W m**-2',
     'Mean top net long-wave radiation flux'),
    ('mean_top_net_long_wave_radiation_flux_clear_sky', 'mtnlwrfcs', 'W m**-2',
     'Mean top net long-wave radiation flux, clear sky'),
    ('mean_top_net_short_wave_radiation_flux', 'mtnswrf', 'W m**-2',
     'Mean top net short-wave radiation flux'),
    ('mean_top_net_short_wave_radiation_flux_clear_sky', 'mtnswrfcs', 'W m**-2',
     'Mean top net short-wave radiation flux, clear sky'),
    ('mean_total_precipitation_rate', 'mtpr', 'kg m**-2 s**-1',
     'Mean total precipitation rate'),
    ('mean_vertical_gradient_of_refractivity_inside_trapping_layer', 'dndza',
     'm**-1', 'Mean vertical gradient of refractivity inside trapping layer'),
    ('mean_vertically_integrated_moisture_divergence', 'mvimd',
     'kg m**-2 s**-1', 'Mean vertically integrated moisture divergence'),
    ('mean_wave_direction', 'mwd', 'Degree true', 'Mean wave direction'),
    ('mean_wave_direction_of_first_swell_partition', 'p140122', 'degrees',
     'Mean wave direction of first swell partition'),
    ('mean_wave_direction_of_second_swell_partition', 'p140125', 'degrees',
     'Mean wave direction of second swell partition'),
    ('mean_wave_direction_of_third_swell_partition', 'p140128', 'degrees',
     'Mean wave direction of third swell partition'),
    ('mean_wave_period', 'mwp', 's', 'Mean wave period'),
    ('mean_wave_period_based_on_first_moment', 'mp1', 's',
     'Mean wave period based on first moment'),
    ('mean_wave_period_based_on_first_moment_for_swell', 'p1ps', 's',
     'Mean wave period based on first moment for swell'),
    ('mean_wave_period_based_on_first_moment_for_wind_waves', 'p1ww', 's',
     'Mean wave period based on first moment for wind waves'),
    ('mean_wave_period_based_on_second_moment_for_swell', 'p2ps', 's',
     'Mean wave period based on second moment for swell'),
    ('mean_wave_period_based_on_second_moment_for_wind_waves', 'p2ww', 's',
     'Mean wave period based on second moment for wind waves'),
    ('mean_wave_period_of_first_swell_partition', 'p140123', 's',
     'Mean wave period of first swell partition'),
    ('mean_wave_period_of_second_swell_partition', 'p140126', 's',
     'Mean wave period of second swell partition'),
    ('mean_wave_period_of_third_swell_partition', 'p140129', 's',
     'Mean wave period of third swell partition'),
    ('mean_zero_crossing_wave_period', 'mp2', 's',
     'Mean zero-crossing wave period'),
    ('medium_cloud_cover', 'mcc', '(0 - 1)', 'Medium cloud cover'),
    ('minimum_vertical_gradient_of_refractivity_inside_trapping_layer', 'dndzn',
     'm**-1',
     'Minimum vertical gradient of refractivity inside trapping layer'),
    ('model_bathymetry', 'wmb', 'm', 'Model bathymetry'),
    ('near_ir_albedo_for_diffuse_radiation', 'alnid', '(0 - 1)',
     'Near IR albedo for diffuse radiation'),
    ('near_ir_albedo_for_direct_radiation', 'alnip', '(0 - 1)',
     'Near IR albedo for direct radiation'),
    ('normalized_energy_flux_into_ocean', 'phioc', 'dimensionless',
     'Normalized energy flux into ocean'),
    ('normalized_energy_flux_into_waves', 'phiaw', 'dimensionless',
     'Normalized energy flux into waves'),
    ('normalized_stress_into_ocean', 'tauoc', 'dimensionless',
     'Normalized stress into ocean'),
    ('northward_gravity_wave_surface_stress', 'mgws', 'N m**-2 s',
     'Northward gravity wave surface stress'),
    ('northward_turbulent_surface_stress', 'nsss', 'N m**-2 s',
     'Northward turbulent surface stress'),
    ('ocean_surface_stress_equivalent_10m_neutral_wind_direction', 'dwi',
     'degrees', '10 metre wind direction'),
    ('ocean_surface_stress_equivalent_10m_neutral_wind_speed', 'wind',
     'm s**-1', '10 metre wind speed'),
    ('orography', 'z', 'm**2 s**-2', 'Geopotential'),
    ('peak_wave_period', 'pp1d', 's', 'Peak wave period'),
    ('period_corresponding_to_maximum_individual_wave_height', 'tmax', 's',
     'Period corresponding to maximum individual wave height'),
    ('potential_evaporation', 'pev', 'm', 'Potential evaporation'),
    ('precipitation_type', 'ptype', 'code table (4.201)', 'Precipitation type'),
    ('runoff', 'ro', 'm', 'Runoff'),
    ('sea_ice_cover', 'siconc', '(0 - 1)', 'Sea ice area fraction'),
    ('sea_surface_temperature', 'sst', 'K', 'Sea surface temperature'),
    ('significant_height_of_combined_wind_waves_and_swell', 'swh', 'm',
     'Significant height of combined wind waves and swell'),
    ('significant_height_of_total_swell', 'shts', 'm',
     'Significant height of total swell'),
    ('significant_height_of_wind_waves', 'shww', 'm',
     'Significant height of wind waves'),
    ('significant_wave_height_of_first_swell_partition', 'p140121', 'm',
     'Significant wave height of first swell partition'),
    ('significant_wave_height_of_second_swell_partition', 'p140124', 'm',
     'Significant wave height of second swell partition'),
    ('significant_wave_height_of_third_swell_partition', 'p140127', 'm',
     'Significant wave height of third swell partition'),
    ('skin_reservoir_content', 'src', 'm of water equivalent',
     'Skin reservoir content'),
    ('skin_temperature', 'skt', 'K', 'Skin temperature'),
    ('slope_of_sub_gridscale_orography', 'slor', '~',
     'Slope of sub-gridscale orography'),
    ('snow_albedo', 'asn', '(0 - 1)', 'Snow albedo'),
    ('snow_density', 'rsn', 'kg m**-3', 'Snow density'),
    ('snow_depth', 'sd', 'm of water equivalent', 'Snow depth'),
    ('snow_evaporation', 'es', 'm of water equivalent', 'Snow evaporation'),
    ('snowfall', 'sf', 'm of water equivalent', 'Snowfall'),
    ('snowmelt', 'smlt', 'm of water equivalent', 'Snowmelt'),
    ('soil_temperature_level_1', 'stl1', 'K', 'Soil temperature level 1'),
    ('soil_temperature_level_2', 'stl2', 'K', 'Soil temperature level 2'),
    ('soil_temperature_level_3', 'stl3', 'K', 'Soil temperature level 3'),
    ('soil_temperature_level_4', 'stl4', 'K', 'Soil temperature level 4'),
    ('soil_type', 'slt', '~', 'Soil type'),
    ('standard_deviation_of_filtered_subgrid_orography', 'sdfor', 'm',
     'Standard deviation of filtered subgrid orography'),
    ('standard_deviation_of_orography', 'sdor', '~',
     'Standard deviation of orography'),
    ('sub_surface_runoff', 'ssro', 'm', 'Sub-surface runoff'),
    ('surface_latent_heat_flux', 'slhf', 'J m**-2', 'Surface latent heat flux'),
    ('surface_net_solar_radiation', 'ssr', 'J m**-2',
     'Surface net solar radiation'),
    ('surface_net_solar_radiation_clear_sky', 'ssrc', 'J m**-2',
     'Surface net solar radiation, clear sky'),
    ('surface_net_thermal_radiation', 'str', 'J m**-2',
     'Surface net thermal radiation'),
    ('surface_net_thermal_radiation_clear_sky', 'strc', 'J m**-2',
     'Surface net thermal radiation, clear sky'),
    ('surface_pressure', 'sp', 'Pa', 'Surface pressure'),
    ('surface_runoff', 'sro', 'm', 'Surface runoff'),
    ('surface_sensible_heat_flux', 'sshf', 'J m**-2',
     'Surface sensible heat flux'),
    ('surface_solar_radiation_downward_clear_sky', 'ssrdc', 'J m**-2',
     'Surface solar radiation downward clear-sky'),
    ('surface_solar_radiation_downwards', 'ssrd', 'J m**-2',
     'Surface solar radiation downwards'),
    ('surface_thermal_radiation_downward_clear_sky', 'strdc', 'J m**-2',
     'Surface thermal radiation downward clear-sky'),
    ('surface_thermal_radiation_downwards', 'strd', 'J m**-2',
     'Surface thermal radiation downwards'),
    ('temperature_of_snow_layer', 'tsn', 'K', 'Temperature of snow layer'),
    ('toa_incident_solar_radiation', 'tisr', 'J m**-2',
     'TOA incident solar radiation'),
    ('top_net_solar_radiation', 'tsr', 'J m**-2', 'Top net solar radiation'),
    ('top_net_solar_radiation_clear_sky', 'tsrc', 'J m**-2',
     'Top net solar radiation, clear sky'),
    ('top_net_thermal_radiation', 'ttr', 'J m**-2',
     'Top net thermal radiation'),
    ('top_net_thermal_radiation_clear_sky', 'ttrc', 'J m**-2',
     'Top net thermal radiation, clear sky'),
    ('total_cloud_cover', 'tcc', '(0 - 1)', 'Total cloud cover'),
    ('total_column_cloud_ice_water', 'tciw', 'kg m**-2',
     'Total column cloud ice water'),
    ('total_column_cloud_liquid_water', 'tclw', 'kg m**-2',
     'Total column cloud liquid water'),
    ('total_column_ozone', 'tco3', 'kg m**-2', 'Total column ozone'),
    ('total_column_rain_water', 'tcrw', 'kg m**-2', 'Total column rain water'),
    ('total_column_snow_water', 'tcsw', 'kg m**-2', 'Total column snow water'),
    ('total_column_supercooled_liquid_water', 'tcslw', 'kg m**-2',
     'Total column supercooled liquid water'),
    ('total_column_water', 'tcw', 'kg m**-2', 'Total column water'),
    ('total_column_water_vapour', 'tcwv', 'kg m**-2',
     'Total column water vapour'),
    ('total_precipitation', 'tp', 'm', 'Total precipitation'),
    ('total_sky_direct_solar_radiation_at_surface', 'fdir', 'J m**-2',
     'Total sky direct solar radiation at surface'),
    ('total_totals_index', 'totalx', 'K', 'Total totals index'),
    ('trapping_layer_base_height', 'tplb', 'm', 'Trapping layer base height'),
    ('trapping_layer_top_height', 'tplt', 'm', 'Trapping layer top height'),
    ('type_of_high_vegetation', 'tvh', '~', 'Type of high vegetation'),
    ('type_of_low_vegetation', 'tvl', '~', 'Type of low vegetation'),
    ('u_component_stokes_drift', 'ust', 'm s**-1', 'U-component stokes drift'),
    ('uv_visible_albedo_for_diffuse_radiation', 'aluvd', '(0 - 1)',
     'UV visible albedo for diffuse radiation'),
    ('uv_visible_albedo_for_direct_radiation', 'aluvp', '(0 - 1)',
     'UV visible albedo for direct radiation'),
    ('v_component_stokes_drift', 'vst', 'm s**-1', 'V-component stokes drift'),
    ('vertical_integral_of_divergence_of_cloud_frozen_water_flux', 'p80.162',
     'kg m**-2 s**-1',
     'Vertical integral of divergence of cloud frozen water flux'),
    ('vertical_integral_of_divergence_of_cloud_liquid_water_flux', 'p79.162',
     'kg m**-2 s**-1',
     'Vertical integral of divergence of cloud liquid water flux'),
    ('vertical_integral_of_divergence_of_geopotential_flux', 'p85.162',
     'W m**-2', 'Vertical integral of divergence of geopotential flux'),
    ('vertical_integral_of_divergence_of_kinetic_energy_flux', 'p82.162',
     'W m**-2', 'Vertical integral of divergence of kinetic energy flux'),
    ('vertical_integral_of_divergence_of_mass_flux', 'p81.162',
     'kg m**-2 s**-1', 'Vertical integral of divergence of mass flux'),
    ('vertical_integral_of_divergence_of_moisture_flux', 'p84.162',
     'kg m**-2 s**-1', 'Vertical integral of divergence of moisture flux'),
    ('vertical_integral_of_divergence_of_ozone_flux', 'p87.162',
     'kg m**-2 s**-1', 'Vertical integral of divergence of ozone flux'),
    ('vertical_integral_of_divergence_of_thermal_energy_flux', 'p83.162',
     'W m**-2', 'Vertical integral of divergence of thermal energy flux'),
    ('vertical_integral_of_divergence_of_total_energy_flux', 'p86.162',
     'W m**-2', 'Vertical integral of divergence of total energy flux'),
    ('vertical_integral_of_eastward_cloud_frozen_water_flux', 'p90.162',
     'kg m**-1 s**-1', 'Vertical integral of eastward cloud frozen water flux'),
    ('vertical_integral_of_eastward_cloud_liquid_water_flux', 'p88.162',
     'kg m**-1 s**-1', 'Vertical integral of eastward cloud liquid water flux'),
    ('vertical_integral_of_eastward_geopotential_flux', 'p73.162', 'W m**-1',
     'Vertical integral of eastward geopotential flux'),
    ('vertical_integral_of_eastward_heat_flux', 'p69.162', 'W m**-1',
     'Vertical integral of eastward heat flux'),
    ('vertical_integral_of_eastward_kinetic_energy_flux', 'p67.162', 'W m**-1',
     'Vertical integral of eastward kinetic energy flux'),
    ('vertical_integral_of_eastward_mass_flux', 'p65.162', 'kg m**-1 s**-1',
     'Vertical integral of eastward mass flux'),
    ('vertical_integral_of_eastward_ozone_flux', 'p77.162', 'kg m**-1 s**-1',
     'Vertical integral of eastward ozone flux'),
    ('vertical_integral_of_eastward_total_energy_flux', 'p75.162', 'W m**-1',
     'Vertical integral of eastward total energy flux'),
    ('vertical_integral_of_eastward_water_vapour_flux', 'p71.162',
     'kg m**-1 s**-1', 'Vertical integral of eastward water vapour flux'),
    ('vertical_integral_of_energy_conversion', 'p64.162', 'W m**-2',
     'Vertical integral of energy conversion'),
    ('vertical_integral_of_kinetic_energy', 'p59.162', 'J m**-2',
     'Vertical integral of kinetic energy'),
    ('vertical_integral_of_mass_of_atmosphere', 'p53.162', 'kg m**-2',
     'Vertical integral of mass of atmosphere'),
    ('vertical_integral_of_mass_tendency', 'p92.162', 'kg m**-2 s**-1',
     'Vertical integral of mass tendency'),
    ('vertical_integral_of_northward_cloud_frozen_water_flux', 'p91.162',
     'kg m**-1 s**-1',
     'Vertical integral of northward cloud frozen water flux '),
    ('vertical_integral_of_northward_cloud_liquid_water_flux', 'p89.162',
     'kg m**-1 s**-1',
     'Vertical integral of northward cloud liquid water flux'),
    ('vertical_integral_of_northward_geopotential_flux', 'p74.162', 'W m**-1',
     'Vertical integral of northward geopotential flux'),
    ('vertical_integral_of_northward_heat_flux', 'p70.162', 'W m**-1',
     'Vertical integral of northward heat flux'),
    ('vertical_integral_of_northward_kinetic_energy_flux', 'p68.162', 'W m**-1',
     'Vertical integral of northward kinetic energy flux'),
    ('vertical_integral_of_northward_mass_flux', 'p66.162', 'kg m**-1 s**-1',
     'Vertical integral of northward mass flux'),
    ('vertical_integral_of_northward_ozone_flux', 'p78.162', 'kg m**-1 s**-1',
     'Vertical integral of northward ozone flux'),
    ('vertical_integral_of_northward_total_energy_flux', 'p76.162', 'W m**-1',
     'Vertical integral of northward total energy flux'),
    ('vertical_integral_of_northward_water_vapour_flux', 'p72.162',
     'kg m**-1 s**-1', 'Vertical integral of northward water vapour flux'),
    ('vertical_integral_of_potential_and_internal_energy', 'p61.162', 'J m**-2',
     'Vertical integral of potential+internal energy'),
    ('vertical_integral_of_potential_internal_and_latent_energy', 'p62.162',
     'J m**-2', 'Vertical integral of potential+internal+latent energy'),
    ('vertical_integral_of_temperature', 'p54.162', 'K kg m**-2',
     'Vertical integral of temperature'),
    ('vertical_integral_of_thermal_energy', 'p60.162', 'J m**-2',
     'Vertical integral of thermal energy'),
    ('vertical_integral_of_total_energy', 'p63.162', 'J m**-2',
     'Vertical integral of total energy'),
    ('vertically_integrated_moisture_divergence', 'vimd', 'kg m**-2',
     'Vertically integrated moisture divergence'),
    ('volumetric_soil_water_layer_1', 'swvl1', 'm**3 m**-3',
     'Volumetric soil water layer 1'),
    ('volumetric_soil_water_layer_2', 'swvl2', 'm**3 m**-3',
     'Volumetric soil water layer 2'),
    ('volumetric_soil_water_layer_3', 'swvl3', 'm**3 m**-3',
     'Volumetric soil water layer 3'),
    ('volumetric_soil_water_layer_4', 'swvl4', 'm**3 m**-3',
     'Volumetric soil water layer 4'),
    ('wave_spectral_directional_width', 'wdw', 'dimensionless',
     'Wave spectral directional width'),
    ('wave_spectral_directional_width_for_swell', 'dwps', 'dimensionless',
     'Wave spectral directional width for swell'),
    ('wave_spectral_directional_width_for_wind_waves', 'dwww', 'dimensionless',
     'Wave spectral directional width for wind waves'),
    ('wave_spectral_kurtosis', 'wsk', 'dimensionless',
     'Wave spectral kurtosis'),
    ('wave_spectral_peakedness', 'wsp', 'dimensionless',
     'Wave spectral peakedness'),
    ('wave_spectral_skewness', 'wss', 'dimensionless',
     'Wave Spectral Skewness'),
    ('zero_degree_level', 'deg0l', 'm', '0 degrees C isothermal level (atm)'),
)
